package cn.edu.sdu.virion.feature.extraction;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import weka.core.Utils;
import cn.edu.sdu.virion.common.Constants;
import cn.edu.sdu.virion.common.Util;
import cn.edu.sdu.virion.core.Sequence;
import cn.edu.sdu.virion.support.database.DBHelper;
import cn.edu.sdu.virion.support.file.FileHelper;
import cn.edu.sdu.virion.support.weka.Classify;

/**
 * ȡ-BiBayes
 * @author Administrator
 *
 */
public class BiBayes {
	
	private Map<String,Integer> countsMap = new HashMap<String,Integer>();
	
	public static void createAllTable()
	{
		String fieldPre = "BiGram_";
		String templateFileName = "CDHIT_SEQS_BIGRAM_XXX.sql";
		for (int i = 1; i<=11; i++) 
			Util.createTable(i, 60+4*(i-1), fieldPre, templateFileName);
	}
	
	@SuppressWarnings("unchecked")
	public static void optimizeParameters() {
		int seqclassIndex = 2;
		int attributeStartIndex = 7;
		for (int i = 0; i < Constants.BIBAYES_HEADFOOT_LEN_COLLECTION.length; i++) {
			String selectTableName = "CDHIT_SEQS_BIGRAM_" + (i + 1);
			String insertTableName = "ATTRS_EVAL";
			int featuresNum = Constants.BIBAYES_HEADFOOT_LEN_COLLECTION[i] * 4;
			int attributeEndIndex = attributeStartIndex + featuresNum - 1;
			String trainSql = String.format(
					"SELECT * FROM %s WHERE STATUS=1 ORDER BY SEQCLASS DESC",
					selectTableName);
			for (String classifierClassName : Constants.WEKA_CLASSIFIERS_CLASS_NAMES) {
				for (int j = 5; j <= 50; j += 5) {
					for (int k = 5; k <= 50; k += 5) {
						String[] removeOptions = Classify.prepareRemoveOptions(
								seqclassIndex, attributeStartIndex,
								attributeEndIndex);
						String[] classifierOptions = new String[6];
						classifierOptions[0] = "-I";
						classifierOptions[1] = String.valueOf(1 * j);
						classifierOptions[2] = "-K";
						classifierOptions[3] = String.valueOf(1 * k);
						classifierOptions[4] = "-S";
						classifierOptions[5] = "1";
						Map result;
						try {
							result = Classify.crossValidate(trainSql,
									removeOptions.clone(), classifierClassName,
									classifierOptions.clone());
							result.put(
									"INFO",
									String.format(
											"BiBayes-HeadFootLen:%d-Classfier:%s-numFeatures:%s-numTrees:%s",
											Constants.BIBAYES_HEADFOOT_LEN_COLLECTION[i],
											classifierClassName,
											classifierOptions[3],
											classifierOptions[1]));
							String insertSql = DBHelper.prepareInsertSql(
									insertTableName, result);
							DBHelper.executeNonQuery(insertSql);
							System.out.println(insertSql);
						} catch (Exception e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}

					}
				}
			}

		}
	}
	
	public static void updateWebSeqs(String jobid)
	{
		Constants.BIBAYES_HEADFOOT_LEN = Constants.BIBAYES_HEADFOOT_LEN_COLLECTION[10];
		try {
			saveFeatures("BiGram", "WEB_SEQS", BiBayes.class.getMethod("orderedFeatures", String.class), String.format("JOBID='%s'", jobid));
		} catch (NoSuchMethodException e) {
			e.printStackTrace();
		} catch (SecurityException e) {
			e.printStackTrace();
		}
	}
	
	public static void updateAll()
	{
		for (int i = 0; i < 11; i++) {
			Constants.BIBAYES_HEADFOOT_LEN = Constants.BIBAYES_HEADFOOT_LEN_COLLECTION[i];
			try {
				saveFeatures("BiGram", "CDHIT_SEQS_BIGRAM_" + (i+1), BiBayes.class.getMethod("orderedFeatures", String.class));
			} catch (NoSuchMethodException e) {
				e.printStackTrace();
			} catch (SecurityException e) {
				e.printStackTrace();
			}
		}
	}
	
	private static void saveFeatures(String fieldName, String tableName, 
			Method m, String strWhere) {
		String sql = String.format("SELECT * FROM %s", tableName);
		if (strWhere != null && strWhere.length()>0) {
			sql += " WHERE " + strWhere;
		}
		Object[] params = new Object[] {};
		ResultSet rs = DBHelper.executeQuery(sql, params);
		try {
			while (rs.next()) {
				Integer id = rs.getInt("ID");
				String sequence = rs.getString("SEQUENCE");
				Object arglist[] = new Object[1];
				arglist[0] = sequence;
				// float[] features = (float[]) m.invoke(m.getClass(), arglist);
				float[] features = orderedFeatures(sequence);
				String valueToSet = Utils.arrayToString(features);
				StringBuilder sb = new StringBuilder();
				for (int i = 1; i <= features.length; i++)
					sb.append(String.format(",%s_%d='%f'", fieldName, i,
							features[i - 1]));
				String insertSql = String.format(
						"UPDATE %s SET status=status%s WHERE ID=%d", tableName,
						sb.toString(), id);
				Object[] insertParams = new Object[] {};
				DBHelper.executeNonQuery(insertSql, insertParams);
			}
		} catch (SQLException | IllegalArgumentException e) {
			e.printStackTrace();
		}
		DBHelper.free(rs);
	}
	
	private static void saveFeatures(String fieldName, String tableName,
			Method m) {
		saveFeatures(fieldName, tableName, m, null);
	}
	
	/**
	 *  BiBayes
	 * @param seq 
	 */
	public static Map<String,Float> extractFeatures(Sequence seq)
	{
		Map<String,Float> resultMap = new HashMap<String,Float>();
		String sequence = seq.getSequence();
		Map positiveMap = DbToMap(Constants.SEQCLASS_POSITIVE);
		Map negativeMap = DbToMap(Constants.SEQCLASS_NEGATIVE);
		String head = sequence.substring(0, Constants.BIBAYES_HEADFOOT_LEN);
		String foot = sequence.substring(sequence.length() - Constants.BIBAYES_HEADFOOT_LEN);
		String headfoot = head + foot;
		for (int i=0; i<headfoot.length(); i++) {
			String key = String.format("%c%d", headfoot.charAt(i), i);
			resultMap.put(Constants.SEQCLASS_POSITIVE + key, (float)positiveMap.get(key));
			resultMap.put(Constants.SEQCLASS_NEGATIVE + key, (float)negativeMap.get(key));
		}
		
		return resultMap;
	}
	
	public static float[] orderedFeatures(String sequence)
	{
		Sequence seq = new Sequence(sequence);
		Map<String,Float> map = extractFeatures(seq);
		float[] result = new float[map.size()];
		String head = sequence.substring(0, Constants.BIBAYES_HEADFOOT_LEN);
		String foot = sequence.substring(sequence.length() - Constants.BIBAYES_HEADFOOT_LEN);
		String headfoot = head + foot;
		for (int i=0; i<headfoot.length(); i++) {
			String key = String.format("%c%d", headfoot.charAt(i), i);
			result[i] = map.get(Constants.SEQCLASS_POSITIVE + key);
			result[Constants.BIBAYES_HEADFOOT_LEN*2 + i] = map.get(Constants.SEQCLASS_NEGATIVE + key);
		}
		return result;
	}
	
	/**
	 * ȡݱBiGramеݲתΪMap
	 * @param seqclass ʶ
	 * @return ƵMap
	 */
	private static Map<String, Float> DbToMap(int seqclass)
	{
		Map<String, Float> map = new HashMap<String, Float>();
		String sql = "SELECT * FROM BiGram WHERE seqclass=?";
		Object[] params = new Object[]{seqclass};
        ResultSet rs = DBHelper.executeQuery(sql, params);
        try {
			while(rs.next()){
				String key = rs.getString("acid") + rs.getString("position");
				Float value = rs.getFloat("frequency");
				map.put(key, value);
			}
		} catch (SQLException e) {
			e.printStackTrace();
		}
        DBHelper.free(rs);
		
		return map;
	}

	/**
	 * BiGramе
	 */
	public void updateBiGram() {
		clearBiGram();
		Constants.BIBAYES_HEADFOOT_LEN = 
				Constants.BIBAYES_HEADFOOT_LEN_COLLECTION[Constants.BIBAYES_HEADFOOT_LEN_COLLECTION.length - 1];
		MapToDb(processSequences(Constants.SEQCLASS_NEGATIVE),
				Constants.SEQCLASS_NEGATIVE);
		MapToDb(processSequences(Constants.SEQCLASS_POSITIVE),
				Constants.SEQCLASS_POSITIVE);
	}
	
	/**
	 * ݱBiGramе
	 */
	private void clearBiGram()
	{
		DBHelper.executeNonQuery("TRUNCATE TABLE BiGram");
	}
	
	/**
	 * ͳݵBiGram
	 * @param map
	 * @param seqclass
	 */
	private void MapToDb(Map map, int seqclass)
	{
		Iterator iterator = map.keySet().iterator();
		String sql = "INSERT INTO BiGram(seqclass,position,acid,frequency) values(?,?,?,?)";
		while(iterator.hasNext()) {
			Object key = iterator.next();
			Object value = map.get(key);
			String temp = (String)key;
			String acid = temp.substring(0,1);
			String position = temp.substring(1);
			Object[] params = new Object[]{seqclass, position, acid, (float)value};
			DBHelper.executeNonQuery(sql, params);
		}
	}
	
	/**
	 * ݿб
	 * @param seqclass ʶ
	 * @return ƵMap
	 */
	private Map processSequences(int seqclass)
	{
		this.initCountsMap();
		int sequenceNum = 0;
		String sql = "SELECT * FROM CDHIT_SEQS WHERE STATUS=1 AND SEQCLASS=?";
        Object[] params = new Object[]{seqclass};
        ResultSet rs = DBHelper.executeQuery(sql, params);
        try {
			while(rs.next()){
				sequenceNum++;
				processSingle(rs.getString("SEQUENCE"));
			}
		} catch (SQLException e) {
			e.printStackTrace();
		}
        DBHelper.free(rs);
		
		return Util.countToFrequency(this.countsMap, sequenceNum);
	}
	
	/**
	 * ƵMapʼ
	 */
	private void initCountsMap()
	{
		this.countsMap.clear();
		for (int i=0; i<Constants.BIBAYES_HEADFOOT_LEN*2; i++) {
			for (int j=0; j<Constants.AMINOACIDS.length; j++) {
				this.fillCountsMap(i, Constants.AMINOACIDS[j]);
			}
		}
	}
	
	/**
	 * Ƶ飬ʼƵԼƵ
	 * @param position λ
	 * @param acid 
	 */
	private void fillCountsMap(int position,Character acid)
	{
		String key = String.format("%c%d", acid, position);
		if (null == this.countsMap.get(key)) {
			this.countsMap.put(key, 0);
		} else {
			int oldValue = this.countsMap.get(key);
			this.countsMap.put(key, oldValue + 1);
		}
	}
	
	/**
	 * ԸƵMap
	 * @param sequence 
	 */
	private void processSingle(String sequence)
	{
		String head = sequence.substring(0, Constants.BIBAYES_HEADFOOT_LEN);
		String foot = sequence.substring(sequence.length() - Constants.BIBAYES_HEADFOOT_LEN);
		String headfoot = head + foot;
		for (int i=0; i<headfoot.length(); i++) {
			fillCountsMap(i, headfoot.charAt(i));
		}
	}
	
}
