package cn.edu.sdu.virion.feature.extraction;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import weka.core.Utils;
import cn.edu.sdu.virion.common.Util;
import cn.edu.sdu.virion.support.database.DBHelper;
import cn.edu.sdu.virion.support.network.HttpRequest;
import cn.edu.sdu.virion.support.weka.Classify;

public class PSEAAC {

	public static void optimizeParameters2() {
		int[] featureNumSets = { 30, 30, 30, 30, 40, 40, 40, 40, 50, 50, 50,
				50, 60, 60, 60, 60 };
		optimizeParameters(featureNumSets, "2");
	}

	public static void optimizeParameters1() {
		int[] featureNumSets = { 25, 25, 25, 25, 30, 30, 30, 30, 35, 35, 35,
				35, 40, 40, 40, 40 };
		optimizeParameters(featureNumSets, "");
	}

	public static void optimizeParameters(int[] featureNumSets, String type) {
		int seqclassIndex = 2;
		int attributeStartIndex = 7;
		String classifierClassName = "weka.classifiers.trees.RandomForest";
		for (int i = 0; i < 16; i++) {
			String selectTableName = "PSEAAC" + type + "_SEQS_" + (i + 1);
			String insertTableName = "ATTRS_EVAL";
			int featuresNum = featureNumSets[i];
			int attributeEndIndex = attributeStartIndex + featuresNum - 1;
			String trainSql = String
					.format("SELECT * FROM %s WHERE STATUS=1 AND USAGE=1 ORDER BY SEQCLASS DESC",
							selectTableName);
			String[] removeOptions = Classify.prepareRemoveOptions(
					seqclassIndex, attributeStartIndex, attributeEndIndex);
			Map result;
			try {
				result = Classify.crossValidate(trainSql,
						removeOptions.clone(), classifierClassName, null);
				result.put("INFO", String.format("PSEAAC%s-%d", type, (i + 1)));
				String insertSql = DBHelper.prepareInsertSql(insertTableName,
						result);
				DBHelper.executeNonQuery(insertSql);
				System.out.println(insertSql);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	public void createAllTable() {
		String fieldPre = "PSEAAC_";
		String templateFileName = "PSEAAC_SEQS_XXX.sql";
		for (int i = 0; i < 4; i++) {
			for (int j = 0; j < 4; j++) {
				Util.createTable(i * 4 + j + 1, 20 + (i + 1) * 5, fieldPre,
						templateFileName);
			}
		}
	}

	public void createAllTable2() {
		String fieldPre = "PSEAAC2_";
		String templateFileName = "PSEAAC2_SEQS_XXX.sql";
		for (int i = 0; i < 4; i++) {
			for (int j = 0; j < 4; j++) {
				Util.createTable(i * 4 + j + 1, 20 + (i + 1) * 10, fieldPre,
						templateFileName);
			}
		}
	}

	public void updateAll() {
		for (int i = 0; i < 4; i++) {
			for (int j = 0; j < 4; j++) {
				String tableName = "PSEAAC_SEQS_" + (i * 4 + j + 1);
				String fieldName = "PSEAAC";
				String weightFactor = String.format("0.%d0", j * 2 + 1);
				String lamda = String.valueOf(5 * (i + 1));
				saveFeatures(tableName, fieldName, weightFactor, lamda);
			}
		}
	}

	public void updateAll2() {
		for (int i = 0; i < 4; i++) {
			for (int j = 0; j < 4; j++) {
				String tableName = "PSEAAC2_SEQS_" + (i * 4 + j + 1);
				String fieldName = "PSEAAC2";
				String weightFactor = String.format("0.%d0", j * 2 + 1);
				String lamda = String.valueOf(5 * (i + 1));
				saveFeatures2(tableName, fieldName, weightFactor, lamda);
			}
		}
	}

	public void saveFeatures(String tableName, String fieldName,
			String weightFactor, String lamda, int type) {
		String sql = String.format(
				"SELECT * FROM %s WHERE STATUS=1 AND PSEAAC2 IS NULL",
				tableName);
		Object[] params = new Object[] {};
		ResultSet rs = DBHelper.executeQuery(sql, params);
		try {
			while (rs.next()) {
				Integer id = rs.getInt("ID");
				String accid = rs.getString("ACC_ID");
				String sequence = rs.getString("SEQUENCE");
				String S1 = String.format(">%s\n%s\n", accid, sequence);
				List<Float> features = batchFeatures(weightFactor, lamda, S1,
						type);
				String valueToSet = Utils.arrayToString(features.toArray());
				StringBuilder sb = new StringBuilder();
				for (int i = 1; i <= features.size(); i++)
					sb.append(String.format(",%s_%d='%f'", fieldName, i,
							features.get(i - 1)));
				String insertSql = String.format(
						"UPDATE %s SET %s='%s'%s WHERE ID=%d", tableName,
						fieldName, valueToSet, sb.toString(), id);
				Object[] insertParams = new Object[] {};
				DBHelper.executeNonQuery(insertSql, insertParams);
				System.out.println(insertSql);
			}
		} catch (SQLException | IllegalArgumentException e) {
			e.printStackTrace();
		}
		DBHelper.free(rs);
	}

	public void saveFeatures2(String tableName, String fieldName,
			String weightFactor, String lamda) {
		this.saveFeatures(tableName, fieldName, weightFactor, lamda, 2);
	}

	public void saveFeatures(String tableName, String fieldName,
			String weightFactor, String lamda) {
		this.saveFeatures(tableName, fieldName, weightFactor, lamda, 1);
	}

	public List<Float> batchFeatures(String T1, String lam, String S1, int type) {
		String url = "http://www.csbio.sjtu.edu.cn/cgi-bin/PseAAC.cgi";
		String paramFormat = "B1=Submit&mode=string&R2=tpse&sushui=sushui&qinshui=qinshui&zhiliang=zhiliang&T1=%s&lam=%s&S1=%s";
		if (type == 2)
			paramFormat = "B1=Submit&mode=string&R2=apse&sushui=sushui&qinshui=qinshui&T1=%s&lam=%s&S1=%s";
		String param = String.format(paramFormat, T1, lam, S1);
		String response = HttpRequest.sendPost(url, param);
		Pattern patternFeature = Pattern.compile("\\d+\\.\\d{3}");
		Matcher mathcerFeature = patternFeature.matcher(response);
		List<Float> result = new ArrayList<Float>();
		while (mathcerFeature.find()) {
			float feature = Float.parseFloat(mathcerFeature.group());
			result.add(feature);
		}
		return result;
	}
}
