package cn.edu.sdu.virion.feature.extraction;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import weka.core.Utils;
import cn.edu.sdu.virion.common.Constants;
import cn.edu.sdu.virion.common.Util;
import cn.edu.sdu.virion.core.Sequence;
import cn.edu.sdu.virion.support.database.DBHelper;
import cn.edu.sdu.virion.support.network.HttpRequest;
import cn.edu.sdu.virion.support.weka.Classify;

/**
 * ȡ-PsePSSM
 * @author Administrator
 *
 */
public class PsePSSM {
	
	public static void createAllTable()
	{
		String fieldPre = "PsePSSM_";
		String templateFileName = "CDHIT_SEQS_PSSM_XXX.sql";
//		Util.createTable(1, 200, fieldPre, templateFileName);
		for (int i = 5 ; i <= 11; i++) {
			Util.createTable(i-4, i*i, fieldPre, templateFileName);
		}
		
	}
	
	public static void optimizeParameters() {
		int seqclassIndex = 2;
		int attributeStartIndex = 7;
		for (int i = 10; i < 17; i++) {
			String selectTableName = "CDHIT_SEQS_PSSM_" + (i < 10 ? 1:("2_"+(i-9)));
			String insertTableName = "ATTRS_EVAL";
			int featuresNum = (i < 10 ? (i+1)*20 : (i-5)*(i-5));
			int attributeEndIndex = attributeStartIndex + featuresNum - 1;
			String trainSql = String.format(
					"SELECT * FROM %s WHERE STATUS=1 ORDER BY SEQCLASS DESC",
					selectTableName);
			for (String classifierClassName : Constants.WEKA_CLASSIFIERS_CLASS_NAMES) {
				for (int j = 5; j <= 50; j += 5) {
					for (int k = 5; k <= 50; k += 5) {
						String[] removeOptions = Classify.prepareRemoveOptions(
								seqclassIndex, attributeStartIndex,
								attributeEndIndex);
						String[] classifierOptions = new String[6];
						classifierOptions[0] = "-I";
						classifierOptions[1] = String.valueOf(1 * j);
						classifierOptions[2] = "-K";
						classifierOptions[3] = String.valueOf(1 * k);
						classifierOptions[4] = "-S";
						classifierOptions[5] = "1";
						Map result;
						try {
							result = Classify.crossValidate(trainSql,
									removeOptions.clone(), classifierClassName,
									classifierOptions.clone());
							result.put(
									"INFO",
									String.format(
											"PSSM-TYPE:%s-Classfier:%s-numFeatures:%s-numTrees:%s",
											i<10 ? "1-"+(i+1):"2-"+(i-9),
											classifierClassName,
											classifierOptions[3],
											classifierOptions[1]));
							String insertSql = DBHelper.prepareInsertSql(
									insertTableName, result);
							DBHelper.executeNonQuery(insertSql);
							System.out.println(insertSql);
						} catch (Exception e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}

					}
				}
			}

		}
	}
	
	
	private static float[] orderedFeatures1(Map<Integer, ArrayList<Float>> map)
	{
		float[] result = new float[Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.length()*Constants.PSSM_LAMDA_MAX];
		int index = 0;
		for (int lamda = Constants.PSSM_LAMDA_MIN; lamda <= Constants.PSSM_LAMDA_MAX; lamda++) {
			ArrayList<Float> array = map.get(lamda);
			for (Float temp : array)
				result[index++] = temp;
 		}
		return result;
	}
	
	private static float[] orderedFeatures2(Map<String, Float> map)
	{
		int groupCount = Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.length();
		float[] result = new float[groupCount*groupCount];
		int index = 0;
		for (int i = 0; i < groupCount; i++) 
			for (int j = 0; j < groupCount; j++) {
				Float temp = map.get(String.format("%c,%c", 
						Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.charAt(i),
						Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.charAt(j)));
				result[index++] = (temp == null ? 0 : temp);
			}
		return result;
	}
	
	private static void updateField(String fieldName) {
		updateField(fieldName, null, null);
	}
	
	private static void updateField(String fieldName, String tableName, String strWhere)
	{		
		String sql;
		if (tableName == "WEB_SEQS" && strWhere != null) {
			sql = String.format(
					"SELECT * FROM %s WHERE %s", 
					tableName, strWhere);
		} else {
			sql = String.format(
					"SELECT * FROM %s WHERE %s IS NULL OR %s=''", 
					Constants.TABLE_NAME_SEQS, fieldName, fieldName);
		}
        Object[] params = new Object[]{};
        ResultSet rs = DBHelper.executeQuery(sql, params);
        try {
			while(rs.next()){
				Integer id = rs.getInt("ID");
				String sequence = rs.getString("SEQUENCE");
				String blast_1_jobid = rs.getString("BLAST_1_JOBID");
				String blast_2_jobid = rs.getString("BLAST_2_JOBID");
				String blast_3_jobid = rs.getString("BLAST_3_JOBID");
				String valueToSet = "";
				if (fieldName.equals("BLAST_1_JOBID")) 
					valueToSet = requestBLAST_1_JOBID(sequence);
				else if (fieldName.equals("BLAST_2_JOBID")) 
					valueToSet = requestBLAST_2_JOBID(blast_1_jobid);
				else if (fieldName.equals("BLAST_3_JOBID")) 
					valueToSet = requestBLAST_3_JOBID(blast_2_jobid);
				else if (fieldName.equals("PsePSSM")) {
					String blast3Result = requestBlast_3_RESULT(blast_3_jobid);
					Map<String, Float> pssmMatrix = blast3ResultToPSSM(blast3Result, sequence);
					Object[] insertParams = new Object[] {};
					
					if (tableName == "WEB_SEQS" && strWhere != null) {
						Constants.PSEPSSM_COMPOSITION_GROUP = Constants.COMPOSITION_GROUP_COLLECTION[6];
						Constants.PSEPSSM_COMPOSITION_GROUP_LEADER = Constants.COMPOSITION_GROUP_LEADER_COLLECTION[6];
						Map<String, Float> condensedMatrix = condensePssm(pssmMatrix, sequence);
						float[] features2 = orderedFeatures2(extractFeature2(condensedMatrix, sequence));
						String valueToSet2 = Utils.arrayToString(features2);
						StringBuilder sb2 = new StringBuilder();
						for (int i = 1; i <= features2.length; i++)
							sb2.append(String.format(",%s_%d='%f'", fieldName, i,
									features2[i - 1]));
						String insertSql2 = String.format(
								"UPDATE %s SET %s='%s'%s WHERE ID=%d", tableName,
								fieldName, valueToSet2, sb2.toString(), id);
						System.out.println(insertSql2);
						DBHelper.executeNonQuery(insertSql2, insertParams);
						valueToSet += valueToSet2;
						
					} else {
						Constants.PSEPSSM_COMPOSITION_GROUP = Constants.COMPOSITION_GROUP_COLLECTION[7];
						Constants.PSEPSSM_COMPOSITION_GROUP_LEADER = Constants.COMPOSITION_GROUP_LEADER_COLLECTION[7];
						float[] features1 = orderedFeatures1(extractFeature(pssmMatrix, sequence));
						String valueToSet1 = Utils.arrayToString(features1);
						StringBuilder sb1 = new StringBuilder();
						for (int i = 1; i <= features1.length; i++)
							sb1.append(String.format(",%s_%d='%f'", fieldName, i,
									features1[i - 1]));
						String insertSql1 = String.format(
								"UPDATE %s SET %s='%s'%s WHERE ID=%d", "CDHIT_SEQS_PSSM_1",
								fieldName, valueToSet1, sb1.toString(), id);
						System.out.println(insertSql1);
						
						DBHelper.executeNonQuery(insertSql1, insertParams);
						valueToSet = valueToSet1;
						
						for (int j = 0; j < 7; j++) {
							Constants.PSEPSSM_COMPOSITION_GROUP = Constants.COMPOSITION_GROUP_COLLECTION[j];
							Constants.PSEPSSM_COMPOSITION_GROUP_LEADER = Constants.COMPOSITION_GROUP_LEADER_COLLECTION[j];
							Map<String, Float> condensedMatrix = condensePssm(pssmMatrix, sequence);
							float[] features2 = orderedFeatures2(extractFeature2(condensedMatrix, sequence));
							String valueToSet2 = Utils.arrayToString(features2);
							StringBuilder sb2 = new StringBuilder();
							for (int i = 1; i <= features2.length; i++)
								sb2.append(String.format(",%s_%d='%f'", fieldName, i,
										features2[i - 1]));
							String insertSql2 = String.format(
									"UPDATE %s SET %s='%s'%s WHERE ID=%d", "CDHIT_SEQS_PSSM_2_" + (j+1),
									fieldName, valueToSet2, sb2.toString(), id);
							System.out.println(insertSql2);
							DBHelper.executeNonQuery(insertSql2, insertParams);
							valueToSet += "|" + valueToSet2;
						}
					}
				}
				String insertSql;
				if (tableName == "WEB_SEQS" && strWhere != null) {
					insertSql = String.format(
							"UPDATE %s SET %s='%s' WHERE ID=%d",
							tableName, fieldName,
							valueToSet, id);
				} else {
					insertSql = String.format(
							"UPDATE %s SET %s='%s' WHERE ID=%d",
							Constants.TABLE_NAME_SEQS, fieldName,
							valueToSet, id);
				}
				Object[] insertParams = new Object[]{};
				DBHelper.executeNonQuery(insertSql, insertParams);
				System.out.println(insertSql);
			}
		} catch (SQLException e) {
			e.printStackTrace();
		}
        DBHelper.free(rs);
	}
	
	
	public static void updateFeaturesByOwn(String tableName, String strWhere) {

		String fieldName = "PsePSSM";
		String sql = String.format("SELECT * FROM %s WHERE %s", tableName, strWhere);
		Object[] params = new Object[] {};
		ResultSet rs = DBHelper.executeQuery(sql, params);
		try {
			while (rs.next()) {
				Integer id = rs.getInt("ID");
				String sequence = rs.getString("SEQUENCE");
				String blast3Result = HttpRequest.sendGet(
						String.format("http://prorep.jnszkj.com/psiblast/run/sequence/%s/id/%d", sequence, id), null);
				Map<String, Float> pssmMatrix = blast3ResultToPSSM(blast3Result, sequence);
				Object[] insertParams = new Object[] {};

				Constants.PSEPSSM_COMPOSITION_GROUP = Constants.COMPOSITION_GROUP_COLLECTION[6];
				Constants.PSEPSSM_COMPOSITION_GROUP_LEADER = Constants.COMPOSITION_GROUP_LEADER_COLLECTION[6];
				Map<String, Float> condensedMatrix = condensePssm(pssmMatrix, sequence);
				float[] features2 = orderedFeatures2(extractFeature2(condensedMatrix, sequence));
				String valueToSet2 = Utils.arrayToString(features2);
				StringBuilder sb2 = new StringBuilder();
				for (int i = 1; i <= features2.length; i++)
					sb2.append(String.format(",%s_%d='%f'", fieldName, i, features2[i - 1]));
				String insertSql2 = String.format("UPDATE %s SET status=status%s WHERE ID=%d", tableName, sb2.toString(), id);
				DBHelper.executeNonQuery(insertSql2, insertParams);
			}
		} catch (SQLException e) {
			e.printStackTrace();
		}
		DBHelper.free(rs);
	}
	
	public void updateFeatures()
	{
		updateField("PsePSSM");
	}	
	
	public void updateBLAST_1_JOBID()
	{
		updateField("BLAST_1_JOBID");
	}
	
	public void updateBLAST_2_JOBID()
	{
		updateField("BLAST_2_JOBID");
	}

	public void updateBLAST_3_JOBID()
	{
		updateField("BLAST_3_JOBID");
	}
	
	public static void updateBLAST_1_JOBID(String tableName, String strWhere) {
		updateField("BLAST_1_JOBID", tableName, strWhere);
	}
	
	public static void updateBLAST_2_JOBID(String tableName, String strWhere) {
		updateField("BLAST_2_JOBID", tableName, strWhere);
	}
	
	public static void updateBLAST_3_JOBID(String tableName, String strWhere) {
		updateField("BLAST_3_JOBID", tableName, strWhere);
	}
	
	public static void updateFeatures(String tableName, String strWhere)
	{
		updateField("PsePSSM", tableName, strWhere);
	}	
	
	
	/**
	 * BLAST_1_JOBID
	 * @param sequence 
	 * @return BLAST_1_JOBID
	 */
	protected static String requestBLAST_1_JOBID(String sequence)
	{
		return HttpRequest.sendPost(
					Constants.URL_PSIBLAST_RUN, 
					String.format("%s&sequence=%s", Constants.URL_PARAM_PSIBLAST, sequence));
	}
	
	/**
	 * JobId
	 * @param previousJobId
	 * @return
	 */
	private static String requestBlastNextJobid(String previousJobId)
	{
		return HttpRequest.sendPost(
				Constants.URL_PSIBLAST_RUN, 
				String.format("%s&previousjobid=%s", Constants.URL_PARAM_PSIBLAST, previousJobId));
	}
	
	protected static String requestBLAST_2_JOBID(String previousJobId)
	{
		return requestBlastNextJobid(previousJobId);
	}
	
	protected static String requestBLAST_3_JOBID(String previousJobId)
	{
		return requestBlastNextJobid(previousJobId);
	}
	
	protected static String requestBlast_3_RESULT(String blast3JobId)
	{
		return HttpRequest.sendGet(
				String.format("%s%s/pssm", Constants.URL_PSIBLAST_RESULT, blast3JobId), null); 
	}
	
	protected static Map<String,Float> blast3ResultToPSSM(String blast3Result, String sequence)
	{		
		Map<String,Float> pssmMatrixMap = new HashMap<String,Float>();
	    Pattern patternLine = Pattern.compile("[A-Z]\\s+([-\\d]+\\s+){20}");
		Matcher mathcerLine = patternLine.matcher(blast3Result);
		int rowsCount = 0;
		while(mathcerLine.find()) {
			String line = mathcerLine.group();
			Pattern patternNumber = Pattern.compile("[-\\d]+");
			Matcher mathcerNumber = patternNumber.matcher(line);
			int colsCount = 0;
			while(mathcerNumber.find()) {
				float value = -1 * Float.parseFloat(mathcerNumber.group());
				value = 1 / (1 + (float) Math.exp(value));
				String key = String.format("%d,%d,%c,%c", 
						colsCount,rowsCount,
						Constants.PSSM_MATRIX_ACID_ORDER.charAt(colsCount),
						sequence.charAt(rowsCount));
				pssmMatrixMap.put(key, value);
				colsCount++;
			}
			rowsCount++;
		}
		return pssmMatrixMap;
	}
	
	private static Map<String,Float> condensePssm(Map<String,Float> pssmMatrix, String sequence)
	{
		Map<String,Float> condensedPssmMatrixMap = new HashMap<String,Float>();
		Map<Character,ArrayList<Integer>> ruleMap = getCondenseRuleMap();
		for (int j = 0; j < sequence.length(); j++) {
			for (int i = 0; i < Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.length(); i++) {
				float value = 0;
				for (Integer index:ruleMap.get(Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.charAt(i))) {
					String oldKey = String.format("%d,%d,%c,%c", index, j, 
							Constants.PSSM_MATRIX_ACID_ORDER.charAt(index),
							sequence.charAt(j));
					value += pssmMatrix.get(oldKey);
				}
					
				String key = String.format("%d,%d,%c,%c", i, j, 
						Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.charAt(i),
						sequence.charAt(j));
				condensedPssmMatrixMap.put(key, value);
			}
		}
		return condensedPssmMatrixMap;
	}
	
	private static Map<Character,ArrayList<Integer>> getCondenseRuleMap()
	{
		Map<Character,ArrayList<Integer>> rulemap = new HashMap<Character,ArrayList<Integer>>();
		for (int i = 0; i < Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.length(); i++) {
			ArrayList<Integer> indice = new ArrayList<Integer>();
			for (int j = 0; j < Constants.PSEPSSM_COMPOSITION_GROUP[i].length(); j++) 
				indice.add(Constants.PSSM_MATRIX_ACID_ORDER.indexOf(Constants.PSEPSSM_COMPOSITION_GROUP[i].charAt(j)));
			rulemap.put(Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.charAt(i), indice);
		}
		return rulemap;
	}

	private static Map<Integer,ArrayList<Float>> extractFeature(Map<String,Float> condensedMatrix, String sequence)
	{
		Map<Integer,ArrayList<Float>> resultMap = new HashMap<Integer,ArrayList<Float>>();
		int seqLen = sequence.length();
		int[] lamdaRange = Util.range(Constants.PSSM_LAMDA_MIN, Constants.PSSM_LAMDA_MAX);
		int groupCount = Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.length();
		for (int k = 0; k < lamdaRange.length; k++){
			int lamda = lamdaRange[k];
			ArrayList<Float> row = new ArrayList<Float>();
			for (int i = 0; i < groupCount; i++)
			{
				float sum = 0;
				for (int j = 0; j < seqLen-lamda-1; j++) {
					String key1 = String.format("%d,%d,%c,%c", i, j, 
							Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.charAt(i),
							sequence.charAt(j));
					String key2 = String.format("%d,%d,%c,%c", i, j+1+lamda, 
							Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.charAt(i),
							sequence.charAt(j+1+lamda));
					float temp1 = condensedMatrix.get(key1);
					float temp2 = condensedMatrix.get(key2);
					sum += Math.pow(temp1-temp2, 2);
				}
				row.add(sum/(seqLen-lamda-1));
			}
			resultMap.put(lamda, row);
		}
		return resultMap;
	}
	
	private static Map<String,Float> extractFeature2(Map<String,Float> condensedMatrix, String sequence)
	{
		String transformedSequence = new Sequence(sequence).transform(
				Constants.PSEPSSM_COMPOSITION_GROUP,Constants.PSEPSSM_COMPOSITION_GROUP_LEADER);
		Map<String,Float> featureMap = new HashMap<String,Float>();
		int seqLen = sequence.length();
		int groupCount = Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.length();
		for (int i = 0; i < groupCount; i++) {
			for (int j = 0; j <seqLen; j++) {
				String newKey = String.format("%c,%c", 
						Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.charAt(i),
						transformedSequence.charAt(j));
				String oldKey = String.format("%d,%d,%c,%c", i, j, 
						Constants.PSEPSSM_COMPOSITION_GROUP_LEADER.charAt(i),
						sequence.charAt(j));
				float value = condensedMatrix.get(oldKey);
				if (null != featureMap.get(newKey)) {
					value += featureMap.get(newKey);
				}
				featureMap.put(newKey, value);
			}
		}
		return featureMap;
	}
	
	
	public static void PsePSSM(Sequence sequence)
	{
		String seq = sequence.getSequence();
		
	}
}
